import importlib as _importlib
if __name__ == "__main__":
    import inspect
    import os
    import sys
    file_path = os.path.dirname(
        os.path.realpath(
            inspect.getfile(
                inspect.currentframe())))
    sys.path.insert(0, os.path.join(file_path, '../'))


_cipher = "origin"
try:
    from ysc2.utils.rsa import helper as _helper
    _new_cipher = _helper.CIPHER
    if _new_cipher:
        _cipher = _new_cipher
    _cipher_class = _importlib.import_module("ysc2.utils.rsa."+_cipher)
except:
    _cipher = "origin"
    _cipher_class = _importlib.import_module("ysc2.utils.rsa.origin")

rsa_utils = _cipher_class.rsa_utils
generate_rsa_keys = _cipher_class.generate_rsa_keys
__all__ = ["rsa_utils","generate_rsa_keys"]


from typing import Tuple, Union, List, Dict, Any
import rsa
import base64
import logging
import rsa
import os
import sys

if __name__ == "__main__":
    import inspect
    file_path = os.path.dirname(
        os.path.realpath(
            inspect.getfile(
                inspect.currentframe())))
    sys.path.insert(0, os.path.join(file_path, '../../'))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG,format="%(asctime)s [%(levelname)s][%(pathname)s:%(lineno)d]: %(message)s")

def generate_rsa_keys(bits=2048)->Tuple:
    pubkey,prikey = rsa.newkeys(bits,poolsize=os.cpu_count())
    return pubkey.save_pkcs1(), prikey.save_pkcs1()

if __name__ == "__main__":
    logging.debug("[RSA][TEST] Generate RSA keys")
    try:
        pubkey,prikey = generate_rsa_keys(bits=2048)
    except:
        import traceback
        logging.error(traceback.format_exc())
        logging.error("[RSA][TEST] Generate RSA keys failed")
    else:
        logging.debug("[RSA][TEST] Generate RSA keys success")
        logging.debug("[RSA][TEST] Public key:")
        logging.debug(str(pubkey,encoding="utf-8"))
        logging.debug("[RSA][TEST] Private key:")
        logging.debug(str(prikey,encoding="utf-8"))

class rsa_utils:
    def __init__(self,pubkey:Union[bytes,str,None]=None,prikey:Union[bytes,str,None]=None):
        self.has_priv_key = False
        self.has_pub_key = False
        self.verify_ok = False
        try:
            if prikey:
                prikey = bytes(prikey)
            self.private_key = rsa.PrivateKey.load_pkcs1(prikey)
        except:
            self.private_key = None
        else:
            self.has_priv_key = True
        try:
            if pubkey:
                pubkey = bytes(pubkey)
            self.public_key = rsa.PublicKey.load_pkcs1(pubkey)
        except:
            self.public_key = None
        else:
            self.has_pub_key = True
        if self.has_priv_key and self.has_pub_key:
            try:
                assert rsa.verify(b"Hello",rsa.sign(b"Hello",self.private_key,"SHA-256"),self.public_key) == "SHA-256"
            except:
                # 打印失败消息
                raise Exception("The Key Pairs are not matched")
            else:
                # 验证通过，设置True
                self.verify_ok = True

    def sign(self,data:bytes)->bytes:
        if self.has_priv_key:
            return rsa.sign(data,self.private_key,"SHA-256")
        else:
            raise Exception("No private key")
    
    def verify(self,data:bytes,sign:bytes)->bool:
        if self.has_pub_key:
            try:
                assert rsa.verify(data,sign,self.public_key) == "SHA-256"
            except rsa.VerificationError:
                return False
            except AssertionError:
                return False
            else:
                return True
        else:
            raise Exception("No public key")

    def encrypt(self,data:bytes)->bytes:
        if self.has_pub_key:
            return rsa.encrypt(data,self.public_key)
        else:
            raise Exception("No public key")

    def decrypt(self,data:bytes)->bytes:
        if self.has_priv_key:
            return rsa.decrypt(data,self.private_key)
        else:
            raise Exception("No private key")

if __name__ == "__main__":
    logging.debug("[RSA][TEST] importing the private key to module.")
    new_rsa_priv = rsa_utils(prikey=prikey)
    logging.debug("[RSA][TEST] Signing data")
    test_sign_bytes = b"""echo "Checking for the latest version of Minecraft Bedrock server..."
curl -H "Accept-Encoding: identity" -H "Accept-Language: en" -L -A "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.33 (KHTML, like Gecko) Chrome/90.0.$RandNum.212 Safari/537.33" -o downloads/version.html https://minecraft.net/en-us/download/server/bedrock/
DownloadURL=$(grep -o 'https://minecraft.azureedge.net/bin-linux/[^"]*' downloads/version.html)
DownloadFile=$(echo "$DownloadURL" | sed 's#.*/##')
echo "$DownloadURL"
echo "$DownloadFile"
# Download latest version of Minecraft Bedrock dedicated server
echo "Downloading the latest version of Minecraft Bedrock server..."
UserName=$(whoami)
curl -H "Accept-Encoding: identity" -H "Accept-Language: en" -L -A "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.33 (KHTML, like Gecko) Chrome/90.0.$RandNum.212 Safari/537.33" -o "downloads/$DownloadFile" "$DownloadURL"
unzip -o "downloads/$DownloadFile" """
    signed = new_rsa_priv.sign(test_sign_bytes)
    logging.debug("[RSA][TEST] Sign data success!")
    logging.debug("[RSA][TEST] importing the public key to module.")
    new_rsa_pub = rsa_utils(pubkey=pubkey)
    logging.debug("[RSA][TEST] Verifying data")
    try:
        assert new_rsa_pub.verify(test_sign_bytes,signed) == True
    except:
        logging.error("[RSA][TEST] Verifying data failed")
    else:
        logging.debug("[RSA][TEST] Verify data success!")

    logging.debug("[RSA][TEST] Encrypting data")
    test_encrypt_bytes = b"Hello"
    encrypted = new_rsa_pub.encrypt(test_encrypt_bytes)
    logging.debug("[RSA][TEST] Encrypt data success!")
    logging.debug("[RSA][TEST] Decrypting data")
    decrypted = new_rsa_priv.decrypt(encrypted)
    try:
        assert decrypted == test_encrypt_bytes
    except:
        logging.error("[RSA][TEST] Decrypt data failed")
    else:
        logging.debug("[RSA][TEST] Decrypt data success!")
    


import os
import sys

if __name__ == "__main__":
    import inspect
    file_path = os.path.dirname(
        os.path.realpath(
            inspect.getfile(
                inspect.currentframe())))
    sys.path.insert(0, os.path.join(file_path, '../../'))


import requests
import json
import logging
import re
from threading import Thread
from ysc2.utils.ipfs import apihelper

if __name__ == "__main__":
    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s [%(levelname)s][%(pathname)s:%(lineno)d]: %(message)s"
    )

def write_to_ipfs(string:str,cid_version=1)->str:
    """
    Write the string to ipfs and return its CID.
    """
    ret = requests.post(url=apihelper.IPFS_API+"/api/v0/add?pin=true&cid-version="+str(cid_version),files={'file':string})
    cid = ret.json()["Hash"]
    return cid

if __name__ == "__main__":
    logging.debug("[IPFS][Test] Testing IPFS")
    test_cid = write_to_ipfs("hello world")

def load_from_ipfs(cid:str)->str:
    """
    Load the content of the cid from ipfs.
    """
    if not re.match("Qm[1-9A-HJ-NP-Za-km-z]{44,}|b[A-Za-z2-7]{58,}|B[A-Z2-7]{58,}|z[1-9A-HJ-NP-Za-km-z]{48,}|F[0-9A-F]{50,}",cid):
        raise Exception("Invalid cid")
    res = requests.get(url=apihelper.IPFS_GW+"/ipfs/"+cid,timeout=60)
    if res.status_code != 200:
        raise Exception("load from IPFS failed")
    return res.text

if __name__ == "__main__":
    test_str = load_from_ipfs(test_cid)
    try:
        assert test_str == "hello world"
    except:
        logging.error("[IPFS][Test] Load String failed")
    else:
        logging.debug("[IPFS][Test] Load String passed")

def download_from_ipfs(cid:str)->bytes:
    """
    Load the content of the cid from ipfs.
    """
    if not re.match("Qm[1-9A-HJ-NP-Za-km-z]{44,}|b[A-Za-z2-7]{58,}|B[A-Z2-7]{58,}|z[1-9A-HJ-NP-Za-km-z]{48,}|F[0-9A-F]{50,}",cid):
        raise Exception("Invalid cid")
    res = requests.get(url=apihelper.IPFS_GW+"/ipfs/"+cid,timeout=60)
    if res.status_code != 200:
        raise Exception("Download failed")
    return res.content

if __name__ == "__main__":
    test_str = download_from_ipfs(test_cid)
    try:
        assert test_str == b"hello world"
    except:
        logging.error("[IPFS][Test] Download failed")
    else:
        logging.debug("[IPFS][Test] Download passed")
    
class pin_to_ipfs(Thread):
    def __init__(self,cid:str,is_background=False):
        Thread.__init__(self)
        self.cid = cid
        self.is_background = is_background
        self.start()
        if not self.is_background:
            self.join()
    def run(self):
        cid = self.cid
        if not re.match("Qm[1-9A-HJ-NP-Za-km-z]{44,}|b[A-Za-z2-7]{58,}|B[A-Z2-7]{58,}|z[1-9A-HJ-NP-Za-km-z]{48,}|F[0-9A-F]{50,}",cid):
            raise Exception("Invalid cid")
        res = requests.post(url=apihelper.IPFS_API+"/api/v0/pin/add?arg="+self.cid,timeout=60)
        if res.status_code != 200:
            logging.error("[IPFS][PIN] Pinning {cid} is Failed, The Response is {res}".format(cid=self.cid,res=res.text))
        else:
            logging.info("[IPFS][PIN] Pinning {cid} is Success".format(cid=self.cid))
        return


import json
import os
import time
import requests
import base64

from ysc2.utils import rsa
from ysc2.utils.rsa import trade as trade_tools
from ysc2.utils.rsa import mutrades as mutrades_tools
from ysc2.utils.ipfs import file as ipfs_file
from ysc2.utils.ipfs import addresses as ipfs_addresses



class ysc:
    def __init__(self,address,privatekey,ysc_gateway_url):
        self.address = address
        self.publickey_str = ipfs_addresses.get_publickey(address)
        self.privatekey = privatekey
        if self.privatekey.startswith("bafkrei"):
            self.privatekey = ipfs_file.download_from_ipfs(self.privatekey)
        self.ysc_gateway_url = ysc_gateway_url
        self.signer = trade_tools.trade_signer(self.publickey_str,self.privatekey)
        r = self.getuserinfo()



    @staticmethod
    def settimezone(timezone):
        os.environ['TZ'] = timezone
        time.tzset()

    def getuserinfo(self):
        url = self.ysc_gateway_url + "/api/v1/getuserinfo"
        data = {"uid":self.address}
        r = requests.post(url,data=data,verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        self.role = ret["role"]
        self.balance = ret["balance"]
        return ret
    
    def getbalance(self):
        url = self.ysc_gateway_url + "/api/v1/getbalance"
        data = {"uid":self.address}
        r = requests.post(url,data=data,verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        self.balance = ret["balance"]
        return ret["balance"]
        
    def getheadblock(self):
        url = self.ysc_gateway_url + "/api/v1/getheadblock"
        r = requests.get(url,verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret
    
    def getblock(self,cid):
        url = self.ysc_gateway_url + "/api/v1/getblock"
        r = requests.post(url,data={"cid":cid},verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret

    def gettrade(self,cid,offset):
        url = self.ysc_gateway_url + "/api/v1/gettrade"
        r = requests.post(url,data={"cid":cid,"offset":offset},verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret

    def getusertrade(self,depth=20):
        url = self.ysc_gateway_url + "/api/v1/getusertrade"
        r = requests.post(url,data={"uid":self.address,"depth":depth},verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret

    def createdeal(self,to,value,transfer_type="Send",content={"cipher":None,"msg":"Transfer"},tax=0,strict_waiting = False):
        transfer = {
            "from": self.address,
            "to": to,
            "type":transfer_type,
            "timestamp":int(time.time()),
            "tax":int(abs(tax)),
            "value":int(value),
            "content":content
        }
        
        signed_transfer = self.signer.sign_trade(transfer)
        json_transfer = json.dumps(signed_transfer)

        r = requests.post(self.ysc_gateway_url+"/api/v1/createdeal",data={"data":str(base64.b64encode(json_transfer.encode()),'utf-8')},verify=False)
        output = r.json()
        if strict_waiting:
            now_block = self.getheadblock()
            if now_block["status"] == False:
                raise Exception(now_block["msg"])
            now_block = now_block["data"]["block"]
            for i in range(0,15):
                time.sleep(10)
                new_block = self.getheadblock()
                if new_block["status"] == False:
                    raise Exception(now_block["msg"])
                new_block_data = new_block["data"]["data"]
                new_block = new_block["data"]["block"]
                if new_block == now_block:
                    continue
                offset = 0
                for j in new_block_data["transfers"]:
                    raw_data = json.loads(base64.b64decode(j["raw_trade"].encode()))
                    if raw_data["sign"] == signed_transfer["sign"]:
                        if j["status"] == 1:
                            return{
                                "status":True,
                                "msg":"success",
                                "block_cid":new_block+"|"+str(offset)
                            }
                        else:
                            return{
                                "status":False,
                                "msg":j["content"]["msg"]
                            }
                    offset+=1
                now_block = new_block
            return{
                "status":False,
                "msg":"timeout"
            }

        if output["status"] == False:
            raise Exception("Create Deal Failed: "+output["msg"])
        else:
            return output

    def createuser(self,rsa_bits=4096,role="user",tax=0,strict_waiting = False):

        newkeys = rsa.generate_rsa_keys(rsa_bits)
        publickey = newkeys[0]
        privatekey = newkeys[1]
        new_publickey_cid = ipfs_file.write_to_ipfs(str(publickey,'utf-8'),cid_version=1)
        new_privatekey_cid = ipfs_file.write_to_ipfs(str(privatekey,'utf-8'),cid_version=1)
        address = ipfs_addresses.upload_pubkey(publickey,role=role)
        msg = str(base64.b64encode(json.dumps({"publickey":new_publickey_cid,"role":role}).encode()),'utf-8')
        transfer = {
            "from": self.address,
            "to": address,
            "type":"AddUser",
            "timestamp":int(time.time()),
            "tax":int(abs(tax)),
            "value":0,
            "content":{
                "cipher":None,
                "msg":msg
            }
        }
        signed_transfer = self.signer.sign_trade(transfer)
        json_transfer = json.dumps(signed_transfer)

        r = requests.post(self.ysc_gateway_url+"/api/v1/createuser",data={"data":str(base64.b64encode(json_transfer.encode()),'utf-8')},verify=False)
        output = r.json()

        if strict_waiting:
            now_block = self.getheadblock()
            if now_block["status"] == False:
                raise Exception(now_block["msg"])
            now_block = now_block["data"]["block"]
            for i in range(0,15):
                time.sleep(10)
                new_block = self.getheadblock()
                if new_block["status"] == False:
                    raise Exception(now_block["msg"])
                new_block_data = new_block["data"]["data"]
                new_block = new_block["data"]["block"]
                if new_block == now_block:
                    continue
                offset = 0
                for j in new_block_data["transfers"]:
                    raw_data = json.loads(base64.b64decode(j["raw_trade"].encode()))

                    if raw_data["sign"] == signed_transfer["sign"]:
                        if j["status"] == 1:
                            return{
                                "status":True,
                                "msg":"success",
                                "block_cid":new_block+"|"+str(offset),
                                "publickey":new_publickey_cid,
                                "privatekey":new_privatekey_cid,
                            }
                        else:
                            return{
                                "status":False,
                                "msg":j["content"]["msg"],
                                "publickey":new_publickey_cid,
                                "privatekey":new_privatekey_cid,
                            }
                    offset+=1
            return{
                "status":False,
                "msg":"timeout",
                "publickey":new_publickey_cid,
                "privatekey":new_privatekey_cid,
                "address":address
            }

        if output["status"] == False:
            raise Exception("Create Account Failed: "+output["msg"])
        else:
            return {
                "publickey":new_publickey_cid,
                "privatekey":new_privatekey_cid,
                "address":address
            }     
class muysc:
    def __init__(self,address,privatekey,ysc_gateway_url):
        self.address = address
        self.publickey_str = ipfs_addresses.get_publickey(address)
        self.privatekey = privatekey
        if self.privatekey.startswith("bafkrei"):
            self.privatekey = ipfs_file.download_from_ipfs(self.privatekey)
        self.ysc_gateway_url = ysc_gateway_url
        self.signer = mutrades_tools.trade_signer(self.address,self.privatekey)
        self.verfier = mutrades_tools.trade_verify(self.address)
        r = self.getuserinfo()



    @staticmethod
    def settimezone(timezone):
        os.environ['TZ'] = timezone
        time.tzset()

    def getuserinfo(self):
        url = self.ysc_gateway_url + "/api/v1/getuserinfo"
        data = {"uid":self.address}
        r = requests.post(url,data=data,verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        self.role = ret["role"]
        self.balance = ret["balance"]
        return ret
    
    def getbalance(self):
        url = self.ysc_gateway_url + "/api/v1/getbalance"
        data = {"uid":self.address}
        r = requests.post(url,data=data,verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        self.balance = ret["balance"]
        return ret["balance"]
        
    def getheadblock(self):
        url = self.ysc_gateway_url + "/api/v1/getheadblock"
        r = requests.get(url,verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret
    
    def getblock(self,cid):
        url = self.ysc_gateway_url + "/api/v1/getblock"
        r = requests.post(url,data={"cid":cid},verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret

    def gettrade(self,cid,offset):
        url = self.ysc_gateway_url + "/api/v1/gettrade"
        r = requests.post(url,data={"cid":cid,"offset":offset},verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret

    def getusertrade(self,depth=20):
        url = self.ysc_gateway_url + "/api/v1/getusertrade"
        r = requests.post(url,data={"uid":self.address,"depth":depth},verify=False)
        ret = r.json()
        if ret["status"] == False:
            raise Exception(ret["msg"])
        return ret

    def createdeal(self,to,value,transfer_type="Send",content={"cipher":None,"msg":"Transfer"},tax=0)->str:
        transfer = {
            "from": self.address,
            "to": to,
            "type":transfer_type,
            "timestamp":int(time.time()),
            "tax":int(abs(tax)),
            "value":int(value),
            "content":content
        }
        signed_transfer = self.signer.sign_trade(transfer)
        return {
            "b64trade":str(base64.b64encode(json.dumps(signed_transfer).encode()),'utf-8')
        }
    
    def createuser(self,rsa_bits=4096,role="user",tax=0,strict_waiting = False)->dict:

        newkeys = rsa.generate_rsa_keys(rsa_bits)
        publickey = newkeys[0]
        privatekey = newkeys[1]
        new_publickey_cid = ipfs_file.write_to_ipfs(str(publickey,'utf-8'),cid_version=1)
        new_privatekey_cid = ipfs_file.write_to_ipfs(str(privatekey,'utf-8'),cid_version=1)
        address = ipfs_addresses.upload_pubkey(publickey,role=role)
        msg = str(base64.b64encode(json.dumps({"publickey":new_publickey_cid,"role":role}).encode()),'utf-8')
        transfer = {
            "from": self.address,
            "to": address,
            "type":"AddUser",
            "timestamp":int(time.time()),
            "tax":int(abs(tax)),
            "value":0,
            "content":{
                "cipher":None,
                "msg":msg
            }
        }
        signed_transfer = self.signer.sign_trade(transfer)
        return {
                "publickey":new_publickey_cid,
                "privatekey_cid":new_privatekey_cid,
                "privatekey":str(privatekey,'utf-8'),
                "address":address,
                "b64trade":str(base64.b64encode(json.dumps(signed_transfer).encode()),'utf-8')
            }

    def sign_transfer(self,b64transfer:str)->str:
        transfer = json.loads(base64.b64decode(b64transfer).decode())
        signed_transfer = self.signer.sign_trade(transfer)
        return str(base64.b64encode(json.dumps(signed_transfer).encode()),'utf-8')

    def merge_sign_transfer(self,b64transfer1:str,b64transfer2:str)->str:
        transfer1 = json.loads(base64.b64decode(b64transfer1).decode())
        transfer2 = json.loads(base64.b64decode(b64transfer2).decode())
        signed_transfer = self.signer.merge_trade(transfer1,transfer2)
        return str(base64.b64encode(json.dumps(signed_transfer).encode()),'utf-8')
    
    def makerequest(self,b64transfer:str,strict_waiting = False):
        signed_transfer = json.loads(base64.b64decode(b64transfer).decode())
        r = requests.post(self.ysc_gateway_url+"/api/v1/createdeal",data={"data":b64transfer},verify=False)
        output = r.json()
        if strict_waiting:
            now_block = self.getheadblock()
            if now_block["status"] == False:
                raise Exception(now_block["msg"])
            now_block = now_block["data"]["block"]
            for i in range(0,15):
                time.sleep(10)
                new_block = self.getheadblock()
                if new_block["status"] == False:
                    raise Exception(now_block["msg"])
                new_block_data = new_block["data"]["data"]
                new_block = new_block["data"]["block"]
                if new_block == now_block:
                    continue
                offset = 0
                for j in new_block_data["transfers"]:
                    raw_data = json.loads(base64.b64decode(j["raw_trade"].encode()))
                    if raw_data["sign"] == signed_transfer["sign"]:
                        if j["status"] == 1:
                            return{
                                "status":True,
                                "msg":"success",
                                "block_cid":new_block+"|"+str(offset)
                            }
                        else:
                            return{
                                "status":False,
                                "msg":j["content"]["msg"]
                            }
                    offset+=1
                now_block = new_block
            return{
                "status":False,
                "msg":"timeout"
            }

        if output["status"] == False:
            raise Exception("Create Deal Failed: "+output["msg"])
        else:
            return output

    
__all__ = ["ysc","muysc"]


import os
import sys

if __name__ == "__main__":
    import inspect
    file_path = os.path.dirname(
        os.path.realpath(
            inspect.getfile(
                inspect.currentframe())))
    sys.path.insert(0, os.path.join(file_path, '../../'))

from ysc2.utils.ipfs import file as ipfs_file

if __name__ == "__main__":
    import logging
    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s [%(levelname)s][%(pathname)s:%(lineno)d]: %(message)s"
    )
    logging.debug("[Address][Test] Generate keys for testing")
    from utils import rsa
    pubkey,privkey = rsa.generate_rsa_keys(bits=2048)
    pubkey_cid = ipfs_file.write_to_ipfs(pubkey,cid_version=1)
    account_addr = "YsA" + pubkey_cid[7:]

def get_publickey(address:str)->bytes:
    if not (address.startswith("YsA") or address.startswith("YsG") or address.startswith("YsF") or address.startswith("YsN") or address.startswith("YsD") or address.startswith("Ysd")):
        raise Exception("Invalid address")
    raw_cid = "bafkrei" + address[3:]
    pubkey = ipfs_file.download_from_ipfs(raw_cid)
    ipfs_file.pin_to_ipfs(raw_cid)
    return pubkey

if __name__ == "__main__":
    logging.debug("[Address][Test] Get the public key of account")
    new_pubkey = get_publickey(account_addr)
    try:
        assert new_pubkey == pubkey
    except:
        logging.error("[Address][Test] Get the public key of account failed")
    else:
        logging.debug("[Address][Test] Get the public key of account success")
    

def upload_pubkey(pubkey:bytes,role="user")->str:
    if (not pubkey.startswith(b"-----BEGIN RSA PUBLIC KEY-----")) and role != "bigmulti":
        raise Exception("Invalid pubkey")
    cid = ipfs_file.write_to_ipfs(pubkey,cid_version=1)
    if role in ["user","admin","superadmin","root"]:
        return "YsA" + cid[7:]
    elif role in ["miner_usergroup"]:
        return "YsG" + cid[7:]
    elif role in ["finance"]:
        return "YsF" + cid[7:]
    elif role in ["blackhole"]:
        return "YsN" + cid[7:]
    elif role in ["multi"]:
        return "Ysd" + cid[7:]
    elif role in ["bigmulti"]:
        return "YsD" + cid[7:]
    else:
        return "YsA" + cid[7:]

if __name__ == "__main__":
    logging.debug("[Address][Test] Upload public key to get the address")
    new_address = upload_pubkey(pubkey,role="User")
    try:
        assert new_address == account_addr
    except:
        logging.error("[Address][Test] Upload public key to get the address failed")
    else:
        logging.debug("[Address][Test] Upload public key to get the address success")

def upload_privkey(privkey:bytes):
    if not privkey.startswith(b"-----BEGIN RSA PRIVATE KEY-----"):
        raise Exception("Invalid privkey")
    cid = ipfs_file.write_to_ipfs(privkey,cid_version=1)
    return cid

if __name__ == "__main__":
    logging.debug("[Address][Test] Upload private key to get the address")
    new_address = upload_privkey(privkey)
    



from ysc2.utils import rsa
import os
import sys
import json
import logging
import base64

class trade_signer(rsa.rsa_utils):
    def __init__(self,*args,**kwargs):
        rsa.rsa_utils.__init__(self,*args,**kwargs)
    
    def sign_trade(self,trade:dict)->dict:
        sorted_trade = dict(sorted(trade.items(), key=lambda x: x[0],reverse=False))
        encoded_trade = json.dumps(sorted_trade)
        signed_trade = base64.b64encode(self.sign(encoded_trade.encode()))
        trade["sign"] = str(signed_trade,encoding="utf-8")
        return trade

    def verify_trade(self,trade:dict)->bool:
        new_trade = trade.copy()
        sign_message = new_trade["sign"]
        sign_message = base64.b64decode(sign_message)
        del new_trade["sign"]
        sorted_trade = dict(sorted(new_trade.items(), key=lambda x: x[0],reverse=False))
        encoded_trade = json.dumps(sorted_trade)
        return self.verify(encoded_trade.encode(),sign_message)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ysc2-sdk",
    version="0.0.4",
    author="Yang",
    author_email="yangruoyu000@live.com",
    description="A simple API for Yeaosound Chain",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.yeaosound.com/yeaosound/ysc/public/ysc-sdk-python",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    extras_require={
        "OpenSSL":"pyOpenSSL"
    },
    install_requires=[
        "requests",
        "rsa"
    ]
)



import os
import sys
import json
import logging
import base64


if __name__ == "__main__":
    import inspect
    file_path = os.path.dirname(
        os.path.realpath(
            inspect.getfile(
                inspect.currentframe())))
    sys.path.insert(0, os.path.join(file_path, '../'))

from ysc2.utils import rsa
from ysc2.utils.ipfs import addresses as ipfs_addresses

def gen_address(bits=2048,num=61):
    addresses = []
    addresses_t = []
    for i in range(num):
        newkeys = rsa.generate_rsa_keys(bits=bits)
        addresses.append(newkeys)
        addresses_t.append(ipfs_addresses.upload_pubkey(newkeys[0],role="multi"))
    raw_address_content = "\n".join(addresses_t)
    raw_address_cid = ipfs_addresses.upload_pubkey(raw_address_content.encode(),role="bigmulti")
    return {
        "address":raw_address_cid,
        "contents":addresses
    }
if __name__ == "__main__":
    new_addresses = gen_address()
class trade_signer():
    def __init__(self,addresses,privatekey):
        self.addresses = addresses # Main Address
        naddress = str(ipfs_addresses.get_publickey(addresses),encoding="utf-8")
        self.naddress = naddress.split("\n") # Each publickey's address
        self.paddress = {} # Each publickey's content
        self.publickeys = {} # Each publickey's library
        self.nprivatekey = privatekey # One privatekey's content
        self.pprivatekey = rsa.rsa_utils(prikey=privatekey) # One privatekey's library
        self.privatekey_from = None # One privatekey's address
        for address in self.naddress:
            paddress = ipfs_addresses.get_publickey(address)
            self.paddress[address] = paddress
            self.publickeys[address] = rsa.rsa_utils(pubkey=paddress)
            if self.publickeys[address].verify(b"Hello",self.pprivatekey.sign(b"Hello")):
                self.privatekey_from = address
                self.publickeys[address] = rsa.rsa_utils(pubkey=paddress,prikey=privatekey)
        if self.privatekey_from is None:
            raise Exception("No privatekey matched")
    def sign_trade(self,trade:dict):
        if "sign" in trade:
            org_sign = trade["sign"]
            org_sign_data = json.loads(base64.b64decode(org_sign))
            del trade["sign"]
        else:
            org_sign_data = {}
        sorted_trade = dict(sorted(trade.items(), key=lambda x: x[0],reverse=False))
        encoded_trade = json.dumps(sorted_trade)
        for addr in org_sign_data:
            if addr not in self.naddress:
                raise Exception("No address matched")
            sign_val = org_sign_data[addr]
            if not self.publickeys[addr].verify(encoded_trade.encode(),base64.b64decode(sign_val)):
                raise Exception("Sign not matched")
        signed_trade = base64.b64encode(self.publickeys[self.privatekey_from].sign(encoded_trade.encode()))
        org_sign_data[self.privatekey_from] = str(signed_trade,encoding="utf-8")
        trade["sign"] = str(base64.b64encode(json.dumps(org_sign_data).encode()),encoding="utf-8")
        return trade
    def merge_trade(self,trade1:dict,trade2:dict):
        if "sign" not in trade1 or "sign" not in trade2:
            raise Exception("No sign in trade")
        org_sign1 = trade1["sign"]
        org_sign_data1 = json.loads(base64.b64decode(org_sign1))
        org_sign2 = trade2["sign"]
        org_sign_data2 = json.loads(base64.b64decode(org_sign2))
        del trade1["sign"]
        del trade2["sign"]
        sorted_trade1 = dict(sorted(trade1.items(), key=lambda x: x[0],reverse=False))
        sorted_trade2 = dict(sorted(trade2.items(), key=lambda x: x[0],reverse=False))
        encoded_trade1 = json.dumps(sorted_trade1)
        encoded_trade2 = json.dumps(sorted_trade2)
        if encoded_trade1 != encoded_trade2:
            raise Exception("Trade not match")
        merge_sign_data = {}
        merge_sign_data.update(org_sign_data1)
        merge_sign_data.update(org_sign_data2)
        merge_sign = base64.b64encode(json.dumps(merge_sign_data).encode())
        trade1["sign"] = str(merge_sign,encoding="utf-8")
        return self.sign_trade(trade1)
        

if __name__ == "__main__":
    original_trade = {
        "from":new_addresses["address"],
        "to":new_addresses["address"],
        "value":1,
        "content":{
            "cipher":None,
            "msg":"Hello"
        },
        "tax":0
    }
    original_trade2 = original_trade.copy()
    signer = []
    for addr in new_addresses["contents"][0:len(new_addresses["contents"])//4+2]:
        signer.append(trade_signer(new_addresses["address"],addr[1]))
        original_trade = signer[-1].sign_trade(original_trade)
    for addr in new_addresses["contents"][len(new_addresses["contents"])//2:len(new_addresses["contents"])//4*3+2]:
        signer.append(trade_signer(new_addresses["address"],addr[1]))
        original_trade2 = signer[-1].sign_trade(original_trade2)
    final_signer = trade_signer(new_addresses["address"],new_addresses["contents"][-1][1])
    final_trade = final_signer.merge_trade(original_trade,original_trade2)
    #signer = trade_signer(new_addresses["address"],new_addresses["contents"][0][1])
    print(json.dumps(final_trade,indent=4))

class trade_verify():
    def __init__(self,addresses):
        self.addresses = addresses # Main Address
        naddress = str(ipfs_addresses.get_publickey(addresses),encoding="utf-8")
        self.naddress = naddress.split("\n") # Each publickey's address
        self.paddress = {} # Each publickey's content
        self.publickeys = {} # Each publickey's library
        for address in self.naddress:
            paddress = ipfs_addresses.get_publickey(address)
            self.paddress[address] = paddress
            self.publickeys[address] = rsa.rsa_utils(pubkey=paddress)

    def verify_trade(self,trade:dict)->bool:
        new_trade = trade.copy()
        sign_message = new_trade["sign"]
        sign_message = base64.b64decode(sign_message)
        del new_trade["sign"]
        sorted_trade = dict(sorted(new_trade.items(), key=lambda x: x[0],reverse=False))
        encoded_trade = json.dumps(sorted_trade)
        len_addrs = len(self.naddress)
        num_of_verify = 0
        org_sign_data = json.loads(sign_message)
        for address in org_sign_data:
            if address in self.naddress:
                if self.publickeys[address].verify(encoded_trade.encode(),base64.b64decode(org_sign_data[address])):
                    num_of_verify += 1
        if num_of_verify > len_addrs//2:
            return True
        else:
            return False

if __name__ == "__main__":
    verify = trade_verify(new_addresses["address"])
    if verify.verify_trade(final_trade):
        print("Trade verified")